(in-package :http-server)


;;; XXX Should abstract this somehow so that it can be done on things
;;; other than filename suffix

(defparameter *content-types*
  '((".html" "text/html")
    (".gif" "image/gif")
    (".jpg" "image/jpeg")
    (".png" "image/png")
    (".class" "application/octet-stream")
    (".doc" "application/octet-stream")
    (".zip" "application/octet-stream")
    (".gz" "application/octet-stream")
    (".ASF" "video/x-ms-asf")
    (".tar" "application/octet-stream")
    (".avi" "video/x-msvideo")
    (".txt" "text/plain")))

(defun read-mime-types (filename)
  "Read a standard-format mime.types file and return an alist suitable for
assigning to *content-types*"
  (labels ((chop-comment (string)
                         (subseq string 0 (position #\# string)))
           (collect-extns (type extns)
                          (loop for e in (split extns)
				if (> (length e) 0)
                                collect (list (s. "." e) type))))
    (with-open-file (in filename :direction :input)
      (let ((mime-types (list))
            (eof (gensym)))
        (loop for line = (read-line in nil eof)
              until (eq line eof)
              for (type extns) =  (remove-if-empty
				   (split (chop-comment line) 2))
              append (collect-extns type extns))))))

(defun copy-stream (from to &optional yield-every yield-proc)
  "Copy into TO from FROM until end of the input file.  Do not
translate or otherwise maul anything.  If supplied, call YIELD-PROC
approximately every YIELD-EVERY bytes."
  (unless (eq (stream-element-type from) (stream-element-type to))
    (error "Incompatible streams ~A and ~A." from to))
  (let ((buf (make-array 4096 :element-type (stream-element-type from)))
        (since-last-pause 0))
    (do ((pos (read-sequence buf from) (read-sequence buf from)))
        ((= 0 pos) nil)
      (if (and yield-every yield-proc)
          (if (> since-last-pause yield-every)
              (progn (setq since-last-pause 0)
                     (funcall yield-proc))
            (incf since-last-pause 4096)))
      (write-sequence buf to :end pos))))

(defun send-file (r file-name &key content-type)
  (let ((stream (request-stream r))
        (if-mod-since (request-if-modified-since r 0)))
    (with-open-file (in file-name :direction :input)
      (peek-char nil in )               ;see if it really is readable
      (cond ((< if-mod-since (file-write-date in))
             (request-send-headers r :content-type content-type
                                   :content-length (file-length in)
                                   :last-modified (file-write-date in))
             (copy-stream in stream))
            (t
             (httpsrv:request-send-headers
              r :response-code 304 :response-text "Not modified"))))))


    

;;; XXX shoul fix to stop people requestiong ../../.. to get out of the
;;; document root.  In practice Apache won't allow it anyway, but better
;;; to be sure
(defun file-request-handler (r arg-string root
                               &optional (directoryindex "index.html"))
  (let* ((file (if (or (eql 0 (length arg-string))
                       (eql (elt arg-string (1- (length arg-string))) #\/))
                   (s. arg-string directoryindex)
                 arg-string))
         (extension (subseq file (or (position #\. file :from-end t)
                                     (length file))))
         (content-type (cadr (assoc extension *content-types* :test #'string=)))
         (fnam (merge-pathnames file root)))
    (if (probe-file fnam)
        (with-open-file (in fnam :direction :input)
          (if (ignore-errors (peek-char nil in nil))
              (send-file r fnam :content-type (or content-type "text/plain"))))
      (request-send-error r 404 (format nil "Can't find file ~S" fnam)))))
